@@ -29,12 +29,14 @@ import quick_server
 import requests
 
 from .util import (
+    ByteResponse,
     df_to_csv,
     get_file_hash,
     get_file_upload_chunk_size,
     get_max_retry,
     get_progress_bar,
     get_retry_sleep,
+    interpret_ctype,
 )
 from .types import (
     CSVBlobResponse,
@@ -72,7 +74,7 @@ else:
     WVD = weakref.WeakValueDictionary
 
 
-__version__ = "0.1.0"
+__version__ = "0.1.1"
 # FIXME: async calls, documentation, auth, summary – time it took etc.
 
 
@@ -195,7 +197,7 @@ class XYMEClient:
             args: Dict[str, Any],
             files: Optional[Dict[str, BytesIO]] = None,
             add_prefix: bool = True,
-            api_version: Optional[int] = None) -> BytesIO:
+            api_version: Optional[int] = None) -> Tuple[BytesIO, str]:
         file_resets = {}
         can_reset = True
         if files is not None:
@@ -301,7 +303,7 @@ class XYMEClient:
             args: Dict[str, Any],
             files: Optional[Dict[str, BytesIO]],
             add_prefix: bool,
-            api_version: Optional[int]) -> BytesIO:
+            api_version: Optional[int]) -> Tuple[BytesIO, str]:
         prefix = ""
         if add_prefix:
             if api_version is None:
@@ -313,7 +315,7 @@ class XYMEClient:
             if req.status_code == 403:
                 raise AccessDenied(req.text)
             if req.status_code == 200:
-                return BytesIO(req.content)
+                return BytesIO(req.content), req.headers["content-type"]
             raise ValueError(
                 f"error {req.status_code} in worker request:\n{req.text}")
         if method == METHOD_POST:
@@ -321,7 +323,7 @@ class XYMEClient:
             if req.status_code == 403:
                 raise AccessDenied(req.text)
             if req.status_code == 200:
-                return BytesIO(req.content)
+                return BytesIO(req.content), req.headers["content-type"]
             raise ValueError(
                 f"error {req.status_code} in worker request:\n{req.text}")
         if method == METHOD_FILE:
@@ -337,7 +339,7 @@ class XYMEClient:
             if req.status_code == 403:
                 raise AccessDenied(req.text)
             if req.status_code == 200:
-                return BytesIO(req.content)
+                return BytesIO(req.content), req.headers["content-type"]
             raise ValueError(
                 f"error {req.status_code} in worker request:\n{req.text}")
         raise ValueError(f"unknown method {method}")
@@ -481,11 +483,11 @@ class XYMEClient:
             args: Dict[str, Any],
             files: Optional[Dict[str, BytesIO]] = None,
             add_prefix: bool = True,
-            api_version: Optional[int] = None) -> BytesIO:
+            api_version: Optional[int] = None) -> Tuple[BytesIO, str]:
         if self._token is None:
             self._login()
 
-        def execute() -> BytesIO:
+        def execute() -> Tuple[BytesIO, str]:
             args["token"] = self._token
             return self._raw_request_bytes(
                 method, path, args, files, add_prefix, api_version)
@@ -660,6 +662,7 @@ class PipelineHandle:
         self._notify_publisher: Optional[str] = None
         self._state: Optional[str] = None
         self._is_high_priority: Optional[bool] = None
+        self._is_parallel: Optional[bool] = None
         self._nodes: Dict[str, NodeHandle] = {}
         self._settings: Optional[Dict[str, Any]] = None
 
@@ -670,6 +673,7 @@ class PipelineHandle:
         self._notify_publisher = None
         self._state = None
         self._is_high_priority = None
+        self._is_parallel = None
         # NOTE: we don't reset nodes
 
     def _maybe_refresh(self) -> None:
@@ -691,6 +695,7 @@ class PipelineHandle:
         self._notify_publisher = info["notify_publisher"]
         self._state = info["state"]
         self._is_high_priority = info["high_priority"]
+        self._is_parallel = info["is_parallel"]
         self._settings = info["settings"]
         old_nodes = {} if self._nodes is None else self._nodes
         self._nodes = {
@@ -742,6 +747,12 @@ class PipelineHandle:
         assert self._is_high_priority is not None
         return self._is_high_priority
 
+    def is_parallel(self) -> bool:
+        self._maybe_refresh()
+        self._maybe_fetch()
+        assert self._is_parallel is not None
+        return self._is_parallel
+
     @contextlib.contextmanager
     def bulk_operation(self) -> Iterator[bool]:
         with self._client.bulk_operation() as do_refresh:
@@ -755,25 +766,52 @@ class PipelineHandle:
     def update_settings(self, settings: Dict[str, Any]) -> None:
         self._client.update_settings(self.get_id(), settings)
 
-    def dynamic(self, input_data: BytesIO) -> BytesIO:
-        cur_res = self._client.request_bytes(
+    def dynamic(self, input_data: BytesIO) -> ByteResponse:
+        cur_res, ctype = self._client.request_bytes(
             METHOD_FILE, "/dynamic", {
                 "pipeline": self._pipe_id,
             }, files={
                 "file": input_data,
-            }).read()
-        if not cur_res:
-            raise ValueError("empty response")
-        return BytesIO(cur_res)
+            })
+        return interpret_ctype(cur_res, ctype)
 
-    def dynamic_obj(self, input_obj: Any) -> Any:
+    def dynamic_obj(self, input_obj: Any) -> ByteResponse:
         bio = BytesIO(json.dumps(
             input_obj,
             separators=(",", ":"),
             indent=None,
             sort_keys=True).encode("utf-8"))
-        out = self.dynamic(bio)
-        return json.load(out)
+        return self.dynamic(bio)
+
+    def dynamic_async(
+            self, input_data: List[BytesIO]) -> List['ComputationHandle']:
+        names = [f"file{pos}" for pos in range(len(input_data))]
+        res: Dict[str, str] = self._client._request_json(
+            METHOD_FILE, "/dynamic_async", {
+                "pipeline": self._pipe_id,
+            }, capture_err=False, files=dict(zip(names, input_data)))
+        return [ComputationHandle(self, res[name]) for name in names]
+
+    def dynamic_async_obj(
+            self, input_data: List[Any]) -> List['ComputationHandle']:
+        return self.dynamic_async([
+            BytesIO(json.dumps(
+                input_obj,
+                separators=(",", ":"),
+                indent=None,
+                sort_keys=True).encode("utf-8"))
+            for input_obj in input_data
+        ])
+
+    def maybe_get_dynamic_result(self, data_id: str) -> Optional[ByteResponse]:
+        cur_res, ctype = self._client.request_bytes(
+            METHOD_GET, "/dynamic_result", {
+                "pipeline": self._pipe_id,
+                "id": data_id,
+            })
+        if not cur_res:
+            return None
+        return interpret_ctype(cur_res, ctype)
 
     def pretty(self, allow_unicode: bool) -> str:
         nodes = [
@@ -1020,6 +1058,7 @@ class NodeHandle:
             "error": "!",
             "unknown": "?",
             "virtual": "∴" if allow_unicode else "V",
+            "parallel": "=",
         }
         return status_map[self.get_status()]
 
@@ -1052,7 +1091,7 @@ class NodeHandle:
             raise ValueError(f"uri is None: {res}")
         return BlobHandle(self._client, uri, is_full=True)
 
-    def read(self, key: str, chunk: int) -> Optional[pd.DataFrame]:
+    def read(self, key: str, chunk: int) -> Optional[ByteResponse]:
         pipeline_id = self.get_pipeline().get_id()
         return self.read_blob(key, chunk).get_content(pipeline_id)
 
@@ -1172,7 +1211,7 @@ class NodeHandle:
                 "key": key,
             }, capture_err=False))
 
-    def get_input_example(self) -> Dict[str, Optional[pd.DataFrame]]:
+    def get_input_example(self) -> Dict[str, Optional[ByteResponse]]:
         if self.get_type() != "custom_data":
             raise ValueError(
                 "can only load example input data for 'custom' node")
@@ -1180,8 +1219,9 @@ class NodeHandle:
         for key in self.get_inputs():
             input_node, out_key = self.get_input(key)
             df = input_node.read(out_key, 0)
-            if df is not None:
-                user_columns = self.get_user_columns(out_key)["user_columns"]
+            if df is not None and isinstance(df, pd.DataFrame):
+                user_columns = \
+                    input_node.get_user_columns(out_key)["user_columns"]
                 rmap = {col: col.replace("user_", "") for col in user_columns}
                 df = df.loc[:, user_columns].rename(columns=rmap)
             res[key] = df
@@ -1390,17 +1430,16 @@ class BlobHandle:
     def get_uri(self) -> str:
         return self._uri
 
-    def get_content(self, pipe_id: str) -> Optional[pd.DataFrame]:
+    def get_content(self, pipe_id: str) -> Optional[ByteResponse]:
         if not self.is_full():
             raise ValueError(f"URI must be full: {self}")
         if self.is_empty():
             return None
-        with self._client._raw_request_bytes(
-                METHOD_POST, "/uri", {
-                    "uri": self._uri,
-                    "pipeline": pipe_id,
-                }) as fin:
-            return pd.read_parquet(fin)
+        fin, ctype = self._client._raw_request_bytes(METHOD_POST, "/uri", {
+            "uri": self._uri,
+            "pipeline": pipe_id,
+        })
+        return interpret_ctype(fin, ctype)
 
     def as_str(self) -> str:
         return f"{self.get_uri()}"
@@ -1425,6 +1464,29 @@ class BlobHandle:
 # *** BlobHandle ***
 
 
+class ComputationHandle:
+    def __init__(self, pipeline: PipelineHandle, data_id: str) -> None:
+        self._pipeline = pipeline
+        self._data_id = data_id
+        self._value: Optional[ByteResponse] = None
+
+    def maybe_value(self) -> Optional[ByteResponse]:
+        if self._value is None:
+            self._value = self._pipeline.maybe_get_dynamic_result(
+                self._data_id)
+        return self._value
+
+    def get_value(self, timeout: Optional[int] = None) -> ByteResponse:
+        start_time = time.monotonic()
+        while timeout is None or (start_time + timeout < time.monotonic()):
+            value = self.maybe_value()
+            if value is not None:
+                return value
+        raise TimeoutError(f"waiting for {self._data_id} has timed out!")
+
+# *** ComputationHandle ***
+
+
 def create_xyme_client(
         url: str,
         user: Optional[str] = None,
